/******************************************************************************
 * Copyright (C) 2010-2011 Lastrix                                            *
 *                                                                            *
 * This file is part of JSONSerialization.                                    *
 *                                                                            *
 * JSONSerialization is free software: you can redistribute it and/or modify  *
 * it under the terms of the GNU General Public License as published by       *
 * the Free Software Foundation, either version 2 of the License, or          *
 * (at your option) any later version.                                        *
 *                                                                            *
 * JSONSerialization is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the               *
 * GNU General Public License for more details.                               *
 *                                                                            *
 * You should have received a copy of the GNU General Public License          *
 * along with JSONSerialization. If not, see <http://www.gnu.org/licenses/>.  *
 ******************************************************************************/

package org.lastrix.JSONSerialization.tool;

import org.json.simple.JSONArray;
import org.json.simple.JSONObject;
import org.json.simple.parser.JSONParser;
import org.json.simple.parser.ParseException;
import org.lastrix.JSONSerialization.Identifiable;
import org.lastrix.JSONSerialization.IdentifiableProvider;
import org.lastrix.JSONSerialization.exception.LoadingFailedException;

import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.nio.CharBuffer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Abstract Loader. Holds all necessary methods for loading objects from json.
 *
 * @author: lastrix
 * Date: 8/13/11
 * Time: 5:33 PM
 */
public abstract class AbstractJSONLoader<T> implements JSONSerializationConstants {

	private final JSONParser parser = new JSONParser();

	/**
	 * Cache required for Identifiable handling
	 */
	private HashMap<Object, Identifiable> cache;

	/**
	 * Creates an empty Loader which could be used for loading from strings.
	 */
	protected AbstractJSONLoader() {
	}

	/// ---------------------------------------------------------------------------------------------- ///
	/// ---------------------------------- Public methods -------------------------------------------- ///
	/// ---------------------------------------------------------------------------------------------- ///

	/**
	 * Load data, return loaded objects as list
	 *
	 * @return list of objects
	 *
	 * @throws LoadingFailedException
	 */
	public final List load() throws LoadingFailedException {
		return load(null);
	}

	/**
	 * Deserialize object from string
	 *
	 * @param <R>        - return type.
	 * @param jsonString - an {@link String} value
	 * @param provider   - the provider
	 *
	 * @return an object of type <R>
	 *
	 * @throws LoadingFailedException
	 */
	public <R> R deserializeObject(final String jsonString, final IdentifiableProvider provider) throws LoadingFailedException {
		final JSONObject root;
		try {
			root = (JSONObject) this.parser.parse(jsonString);
		} catch (ParseException e) {
			throw new LoadingFailedException(e);
		}
		resetCache();
		return (R) this.fromJSON(root.containsKey(DATA_TAG) ? root.get(DATA_TAG) : root, provider);
	}

	/**
	 * Deserialize object from string
	 *
	 * @param <R>        - return type.
	 * @param jsonString - an {@link String} value
	 *
	 * @return an object of type <R>
	 *
	 * @throws LoadingFailedException
	 */
	public <R> R deserializeObject(final String jsonString) throws LoadingFailedException {
		return deserializeObject(jsonString, null);
	}

	/**
	 * Deserialize objects from string
	 *
	 * @param json     - the json string
	 * @param provider - the provider
	 *
	 * @return an List of objects
	 *
	 * @throws LoadingFailedException
	 */
	public final List deserialize(final String json, final IdentifiableProvider provider) throws LoadingFailedException {
		final StringReader reader = new StringReader(json);
		final List list = deserialize(reader, provider);
		reader.close();
		return list;
	}

	/**
	 * Deserialize objects from string
	 *
	 * @param json - the json string
	 *
	 * @return an List of objects
	 *
	 * @throws LoadingFailedException
	 */
	public final List deserialize(final String json) throws LoadingFailedException {
		return deserialize(json, null);
	}

	/// ---------------------------------------------------------------------------------------------- ///
	/// ---------------------------------- Private methods ------------------------------------------- ///
	/// ---------------------------------------------------------------------------------------------- ///

	/**
	 * Convert json objects to valid user object
	 *
	 * @param o        - the JSONObject or like
	 * @param provider - the provider
	 *
	 * @return an object
	 *
	 * @throws LoadingFailedException
	 */
	private Object fromJSON(final Object o, final IdentifiableProvider provider) throws LoadingFailedException {
		if (o instanceof JSONObject) {

			final JSONObject json = (JSONObject) o;
			final Class<?> clazz;
			try {
				clazz = Class.forName((String) json.get(TYPE_INFO));
			} catch (ClassNotFoundException e) {
				throw new LoadingFailedException(e);
			}

			if (clazz.isArray()) {
				//array handler
				return arrayFromJSON(json, provider);

			} else if (clazz.isEnum()) {
				// enums are easy
				return Enum.valueOf(
				                   (Class<? extends Enum>) clazz,
				                   (String) json.get(ENUM_VALUE)
				                   );
			}

			Object r = null;
			Object id = null;
			if (this.isImplementsIdentifiable(clazz)) {
				id = this.fromJSON(((JSONObject) json.get(FIELDS)).get(ID), provider);
				r = getIdentifiable(provider, id);
				if (r == null) {
					r = getFromCache(id);
				}
			}

			if (r == null) {
				r = createInstance(clazz, json, provider);
			}

			if (r instanceof Identifiable) {
				((Identifiable) r).setId(id);
				loaded((Identifiable) r);
			}

			// so we've got an object
			if (r instanceof Map) {
				final JSONObject map = (JSONObject) json.get(MAP);
				for (final Object key : map.keySet()) {
					((Map) r).put(this.fromJSON(key, provider), this.fromJSON(map.get(key), provider));
				}
			} else if (r instanceof List) {
				final JSONArray array = (JSONArray) json.get(LIST);
				for (final Object value : array) {
					((List) r).add(this.fromJSON(value, provider));
				}
			} else {
				final JSONObject fields = (JSONObject) json.get(FIELDS);
				// it's damn hell
				Class<?> rc = r.getClass();
				while (rc != null) {
					for (final Field field : rc.getDeclaredFields()) {
						if (((field.getModifiers() & Modifier.FINAL) == 0)) {
							// we can not access that class
							field.setAccessible(true);
							if (fields.containsKey(field.getName())) {
								try {
									field.set(r, this.fromJSON(fields.get(field.getName()), provider));
								} catch (IllegalAccessException e) {
									throw new LoadingFailedException(e);
								}
							}
						}
					}
					rc = rc.getSuperclass();
				}
			}
			return r;
		} else {
			// this is simple type (java.lang.* or null)
			if (o instanceof String) {
				final String string = (String) o;
				if (string.equals(NULL_OBJECT)) {
					return null;
				} else if (string.startsWith(OBJECT_LINK)) {
					// L: means - link to (id)
					// without identifiableObjectProvider we cann't get our object
					final Object simple = this.createSimple(string.substring(2));
					// if object is simple just return it
					if (simple != null) { return getIdentifiable(provider, simple); }
					// and worst case - if this thing is real JSON formed object that damn JSONSimple is not deserialized since it
					// cann't allow non-string keys
					try {
						return getIdentifiable(provider, fromJSON(parser.parse(string.substring(2)), provider));
					} catch (ParseException e) {
						throw new LoadingFailedException(e);
					}
				} else {
					final Object result = this.createSimple(string);
					if (result != null) {
						return result;
					} else {
						try {
							return fromJSON(parser.parse(string), provider);
						} catch (ParseException e) {
							throw new LoadingFailedException(e);
						}
					}
				}
			}
		}
		throw new IllegalArgumentException("Illegal parameter : " + o);
	}

	/**
	 * Safely return Identifiable from provider ( if possible ), or from cache
	 *
	 * @param provider - the provider
	 * @param id       - the id
	 *
	 * @return Identifiable or null
	 */
	private Identifiable getIdentifiable(final IdentifiableProvider provider, Object id) {
		if (provider == null) {
			return getFromCache(id);
		}

		final Identifiable o = provider.get(id);
		if (o == null) {
			return getFromCache(id);
		}
		return o;
	}

	/**
	 * Create instance of class
	 *
	 * @param clazz    - the class
	 * @param json     - object that holds object information
	 * @param provider - the provider
	 *
	 * @return - an Object
	 *
	 * @throws LoadingFailedException
	 */
	private Object createInstance(final Class<?> clazz, final JSONObject json, final IdentifiableProvider provider) throws LoadingFailedException {
		try {
			Object r;
			if (clazz.getEnclosingClass() == null) {
				final Constructor<?> c = clazz.getDeclaredConstructor();

				// to allow private constructors usage.
				if (!c.isAccessible()) {
					c.setAccessible(true);
				}
				r = c.newInstance();
			} else {
				// try the damn ugly way
				final Field parent = this.getNestedClassParentField(clazz);
				if (parent == null) {
					final Constructor<?> c = clazz.getDeclaredConstructor();

					// to allow private constructors usage.
					if (!c.isAccessible()) {
						c.setAccessible(true);
					}
					r = c.newInstance();
				} else {
					final Constructor<?> c = clazz.getDeclaredConstructor(clazz.getEnclosingClass());

					// to allow private constructors usage.
					if (!c.isAccessible()) {
						c.setAccessible(true);
					}
					r = c.newInstance(getIdentifiable(provider, this.fromJSON(json.get(ENCLOSING_ID), provider)));
				}
			}
			return r;
		} catch (Exception e) {
			throw new LoadingFailedException(e);
		}
	}

	/**
	 * Create array from json
	 *
	 * @param json     - the json representation
	 * @param provider - the provider
	 *
	 * @return an Array
	 *
	 * @throws LoadingFailedException
	 */
	private Object arrayFromJSON(final JSONObject json, final IdentifiableProvider provider) throws LoadingFailedException {
		// we cann't call constructor of an array, so do the valid array creation manually.
		final JSONArray array = (JSONArray) json.get(ARRAY);
		final long aSize = (Long) json.get(ARRAY_SIZE);
		final Object a = Array.newInstance(findClassByName((String) json.get(ARRAY_COMPONENT)), (int) aSize);
		int i = 0;
		for (final Object value : array) {
			Array.set(a, i++, this.fromJSON(value, provider));
		}
		return a;
	}

	/// ---------------------------------------------------------------------------------------------- ///
	/// -------------------------------- Protected methods ------------------------------------------- ///
	/// ---------------------------------------------------------------------------------------------- ///

	/**
	 * Load objects from reader
	 *
	 * @param reader - the reader
	 *
	 * @return an List of objects
	 *
	 * @throws LoadingFailedException
	 */
	protected List deserialize(final Reader reader) throws LoadingFailedException {
		return deserialize(reader, null);
	}

	/**
	 * Load objects from reader
	 *
	 * @param reader   - the reader
	 * @param provider - the provider
	 *
	 * @return an List of objects
	 *
	 * @throws LoadingFailedException
	 */
	protected List deserialize(final Reader reader, final IdentifiableProvider provider) throws LoadingFailedException {
		final CharBuffer cb = CharBuffer.allocate(1024);
		String text = "";
		List list = new ArrayList();

		resetCache();

		boolean read;
		try {
			while (reader.ready()) {
				// feed buffer till we hit a DELIMITER
				read = true;
				while (!text.contains(DELIMITER)) {
					if (reader.read(cb) <= 0) {
						read = false;
						break;
					}
					cb.flip();
					text += cb.toString();
					cb.clear();
					if (!reader.ready()) { break; }
				}
				if (!read && text.isEmpty()) {
					break;
				}
				int iod = 0;
				int ciod;
				// run trough string and cut probable objects
				while ((ciod = text.indexOf(DELIMITER, iod)) != -1) {
					//store in list
					try {
						list.add(fromJSON(((JSONObject) parser.parse(text.substring(iod, ciod))).get(DATA_TAG), provider));
					} catch (ParseException e) {
						throw new LoadingFailedException(e);
					}
					iod = ciod + DELIMITER.length();
				}
				// if something missed add it to text
				if (iod < text.length()) {
					text = text.substring(iod);
				} else {
					// or clear it
					text = "";
				}
			}
		} catch (IOException e) {
			throw new LoadingFailedException(e);
		}
		return list;
	}

	/**
	 * Add Identifiable to cache
	 *
	 * @param o - an Identifiable
	 */
	protected final void loaded(Identifiable o) {
		cache.put(o.getId(), o);
	}

	/**
	 * Return identifiable from cache by id
	 *
	 * @param id - the id
	 *
	 * @return Identifiable or null
	 */
	protected final Identifiable getFromCache(Object id) {
		return cache.get(id);
	}

	/**
	 * Resets cache
	 */
	protected void resetCache() {
		if (this.cache == null) {
			this.cache = new HashMap<Object, Identifiable>();
		} else {
			this.cache.clear();
		}
	}

	/**
	 * Returns true if {@link Identifiable} interface implemented by this class
	 *
	 * @param clazz - an {@link Class} object
	 *
	 * @return an {@link Boolean} value
	 */
	protected final static boolean isImplementsIdentifiable(final Class<?> clazz) {
		//FIXME: find better way
		for (final Class<?> c : clazz.getInterfaces()) {
			if (c.equals(Identifiable.class)) { return true; }
		}
		return false;
	}

	/**
	 * @param clazz - an {@link Class} object
	 *
	 * @return an {@link Field} object
	 */
	protected final static Field getNestedClassParentField(final Class<?> clazz) {
		Field parent = null;
		for (final Field f : clazz.getDeclaredFields()) {
			if (f.getType().equals(clazz.getEnclosingClass())) {
				parent = f;
				break;
			}
		}
		return parent;
	}

	/**
	 * Find class by its short name (only for primitive types)
	 *
	 * @param name - the name of class
	 *
	 * @return
	 */
	protected final static Class<?> findClassByName(String name) {
		try {
			return Class.forName(name);
		} catch (Exception e) {
			if (Long.class.getName().equals(name)) {
				return Long.class;
			} else if (Integer.class.getName().equals(name)) {
				return Integer.class;
			} else if (Byte.class.getName().equals(name)) {
				return Byte.class;
			} else if (Double.class.getName().equals(name)) {
				return Double.class;
			} else if (Float.class.getName().equals(name)) {
				return Float.class;
			} else if (Boolean.class.getName().equals(name)) {
				return Boolean.class;
			}
			return null;
		}
	}

	/**
	 * Creates simple type that cann't be serialized properly die to infinite loops (damn integer class has his own fields of type
	 * integer...)
	 *
	 * @param string - an {@link String} value
	 *
	 * @return an object or null
	 */
	protected final static Object createSimple(final String string) {
		final String type = string.substring(0, string.indexOf(":"));
		final String value = string.substring(string.indexOf(":") + 1);
		final Class clazz;
		try {
			clazz = Class.forName(type);
		} catch (ClassNotFoundException e) {
			return null;
		}
		if (String.class.equals(clazz)) {
			return value;
		} else if (Integer.class.equals(clazz)) {
			return Integer.parseInt(value);
		} else if (Long.class.equals(clazz)) {
			return Long.parseLong(value);
		} else if (Byte.class.equals(clazz)) {
			return Byte.parseByte(value);
		} else if (Float.class.equals(clazz)) {
			return Float.parseFloat(value);
		} else if (Double.class.equals(clazz)) {
			return Double.parseDouble(value);
		} else if (Boolean.class.equals(clazz)) { return Boolean.parseBoolean(value); }
		return null;
	}


	/// ---------------------------------------------------------------------------------------------- ///
	/// --------------------------------- Abstract methods ------------------------------------------- ///
	/// ---------------------------------------------------------------------------------------------- ///


	/**
	 * Load data, return loaded objects as list
	 *
	 * @param provider - the provider that possibly could return us identifiable for an id, so you may make an update
	 *
	 * @return list of objects
	 *
	 * @throws LoadingFailedException
	 */
	public abstract List load(final IdentifiableProvider<T> provider) throws LoadingFailedException;
}
