/******************************************************************************
 * Copyright (C) 2010-2011 Lastrix                                            *
 *                                                                            *
 * This file is part of JSONSerialization.                                    *
 *                                                                            *
 * JSONSerialization is free software: you can redistribute it and/or modify  *
 * it under the terms of the GNU General Public License as published by       *
 * the Free Software Foundation, either version 2 of the License, or          *
 * (at your option) any later version.                                        *
 *                                                                            *
 * JSONSerialization is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the               *
 * GNU General Public License for more details.                               *
 *                                                                            *
 * You should have received a copy of the GNU General Public License          *
 * along with JSONSerialization. If not, see <http://www.gnu.org/licenses/>.  *
 ******************************************************************************/

package org.lastrix.JSONSerialization.tool;

import org.json.simple.JSONArray;
import org.json.simple.JSONObject;
import org.lastrix.JSONSerialization.DefaultPolicy;
import org.lastrix.JSONSerialization.Identifiable;
import org.lastrix.JSONSerialization.Policy;
import org.lastrix.JSONSerialization.exception.SavingFailedException;

import java.io.StringWriter;
import java.io.Writer;
import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.*;

/**
 * Basic class for saving object into JSON. Holds all necessary methods for serialization.
 *
 * @author: lastrix
 * Date: 8/13/11
 * Time: 4:51 PM
 * @see #serialize(Collection)
 * @see #serializeObject(Object)
 * @see #serialize(Writer, Collection)
 */
public abstract class AbstractJSONSaver implements JSONSerializationConstants {

	/**
	 * Cache required for saving only one copy of Identifiable
	 */
	private HashMap<Object, Identifiable> cache;

	/**
	 * Create default saver.
	 */
	protected AbstractJSONSaver() {
	}


	/// ---------------------------------------------------------------------------------------------- ///
	/// ---------------------------------- Public methods -------------------------------------------- ///
	/// ---------------------------------------------------------------------------------------------- ///

	/**
	 * Save objects
	 *
	 * @param objects - the objects container
	 *
	 * @throws SavingFailedException
	 */
	public final void save(final Collection<?> objects) throws SavingFailedException {
		save(objects, null);
	}

	/**
	 * Convert object into json string
	 *
	 * @param o      - an object
	 * @param policy - an {@link Policy} object
	 *
	 * @return an {@link String} value
	 *
	 * @throws SavingFailedException
	 */
	public final String serializeObject(final Object o, final Policy policy) throws SavingFailedException {
		return serialize(Collections.singletonList(o), policy);
	}

	/**
	 * Convert object into json string
	 *
	 * @param o - an object
	 *
	 * @return an {@link String} value
	 *
	 * @throws SavingFailedException
	 */
	public final String serializeObject(final Object o) throws SavingFailedException {
		return serializeObject(o, null);
	}

	/**
	 * Convert collection of objects into json string
	 *
	 * @param collection - the objects
	 * @param policy     - the saving policy
	 *
	 * @return and String
	 *
	 * @throws SavingFailedException
	 */
	public final String serialize(final Collection<?> collection, final Policy policy) throws SavingFailedException {
		final StringWriter sw = new StringWriter();
		serialize(sw, collection, policy);
		sw.flush();
		return sw.getBuffer().toString();
	}

	/**
	 * Convert collection of objects into json string
	 *
	 * @param collection - the objects
	 *
	 * @return and String
	 *
	 * @throws SavingFailedException
	 */
	public final String serialize(final Collection<?> collection) throws SavingFailedException {
		return serialize(collection);
	}

	/// ---------------------------------------------------------------------------------------------- ///
	/// ---------------------------------- Private methods ------------------------------------------- ///
	/// ---------------------------------------------------------------------------------------------- ///


	/**
	 * Recursively convert object into JSON notation
	 *
	 * @param o      - an {@link Object}
	 * @param policy - an {@link Policy} object
	 *
	 * @return an {@link Object}
	 *
	 * @throws SavingFailedException
	 */
	private Object toJSON(final Object o, final Policy policy) throws SavingFailedException {
		// check special cases
		//nulls always needed
		if (o == null) {
			return NULL_OBJECT;
		} else if (isSimple(o)) {
			return o.getClass().getName() + ":" + o;
		}

		// okay, our object is complex
		if (o.getClass().isAnonymousClass() || o.getClass().isLocalClass()) {
			throw new IllegalArgumentException(String.format("'%s' can not be serialized (Anon or local class).", o));
		}
		if (o.getClass().isEnum()) {
			// enum handler
			return enumToJSON(o);

		} else if (o.getClass().isArray()) {
			// array handler
			return arrayToJSON(o, policy);

		} else if (o instanceof Map) {
			// map handler
			return mapToJSON((Map) o, policy);

		} else if (o instanceof List) {
			// list handler
			return listToJSON((List) o, policy);

		}
		return objectToJSON(o, policy);
	}

	/**
	 * Convert enum to json
	 *
	 * @param o - an enum constant
	 *
	 * @return an JSONObject
	 */
	private static final JSONObject enumToJSON(final Object o) {
		final JSONObject json = new JSONObject();
		// put type info for future object creation
		json.put(TYPE_INFO, o.getClass().getName());
		json.put(ENUM_VALUE, o.toString());
		return json;
	}

	/**
	 * Convert object to json
	 *
	 * @param o      - the object
	 * @param policy - the policy to determine which fields are needed
	 *
	 * @return an JSONObject or String or empty JSONObject
	 *
	 * @throws SavingFailedException
	 */
	private Object objectToJSON(final Object o, final Policy policy) throws SavingFailedException {
		if (o instanceof Identifiable && isSaved((Identifiable) o)) {
			//if identifiable already saved we should return link to it
			return makeLink((Identifiable) o, policy);
		}

		final JSONObject json = new JSONObject();

		// put type info for future object creation
		json.put(TYPE_INFO, o.getClass().getName());

		//check for enclosing types, we can not save object if parent is not Identifiable
		if (o.getClass().getEnclosingClass() != null) {
			final Field parent = getNestedClassParentField(o.getClass());
			if (parent != null) {
				if (!isImplementsIdentifiable(o.getClass().getEnclosingClass())) {
					throw new IllegalArgumentException(String.format("Object '%s' can not be saved: Enclosing type is not Identifiable."));
				}

				if (!parent.isAccessible()) { parent.setAccessible(true); }
				try {
					json.put(ENCLOSING_ID, this.toJSON(((Identifiable) parent.get(o)).getId(), policy));
				} catch (IllegalAccessException e) {
					throw new SavingFailedException(e);
				}
			}
		}

		final JSONObject jsonFields = new JSONObject();
		json.put(FIELDS, jsonFields);

		// do the damn field fetching...
		Class<?> oc = o.getClass();
		while (oc != null) {
			for (final Field field : oc.getDeclaredFields()) {
				if ((field.getModifiers() & Modifier.FINAL) == 0) {
					field.setAccessible(true);
					Object value = null;
					try {
						value = field.get(o);
					} catch (IllegalAccessException e) {
						//send it to higher level
						throw new SavingFailedException(e);
					}
					switch (policy.isNeeded(o, field)) {
						case AS_LINK:
							if (value instanceof Identifiable) {
								jsonFields.put(field.getName(), makeLink((Identifiable) value, policy));
							}
							break;

						case YES:
							jsonFields.put(field.getName(), this.toJSON(value, policy));
							break;

						case NO:
							break;
					}
				}
			}
			oc = oc.getSuperclass();
		}

		//put info to cache that we have saved identifiable
		if (o instanceof Identifiable) {
			saved((Identifiable) o);
		}
		return json;
	}

	/**
	 * Create link to specified Identifiable
	 *
	 * @param o      - the identifiable
	 * @param policy - the policy
	 *
	 * @return an String
	 *
	 * @throws SavingFailedException
	 */
	private final String makeLink(final Identifiable o, final Policy policy) throws SavingFailedException {
		return OBJECT_LINK + toJSON(o.getId(), policy);
	}

	/**
	 * Convert list to json
	 *
	 * @param o      - the list
	 * @param policy - policy to determine which items are needed
	 *
	 * @return an JSONObject
	 *
	 * @throws SavingFailedException
	 */
	private final JSONObject listToJSON(final List o, final Policy policy) throws SavingFailedException {
		final JSONObject json = new JSONObject();
		// put type info for future object creation
		json.put(TYPE_INFO, o.getClass().getName());
		final JSONArray array = new JSONArray();
		json.put(LIST, array);
		for (final Object value : (List) o) {
			switch (policy.isNeeded(value)) {
				case AS_LINK:
					if (value instanceof Identifiable) {
						array.add(makeLink((Identifiable) value, policy));
					}
					break;

				case YES:
					array.add(this.toJSON(value, policy));
					break;

				case NO:
					break;
			}
		}
		return json;
	}

	/**
	 * Convert map to json
	 *
	 * @param o      - the map
	 * @param policy - policy to determine which values should be stored
	 *
	 * @return JSONObject
	 *
	 * @throws SavingFailedException
	 */
	private final JSONObject mapToJSON(final Map o, final Policy policy) throws SavingFailedException {

		// if object is Map we should only fetch all keys and values
		final JSONObject json = new JSONObject();
		// put type info for future object creation
		json.put(TYPE_INFO, o.getClass().getName());

		final JSONObject jsonMap = new JSONObject();
		json.put(MAP, jsonMap);
		for (final Object key : ((Map) o).keySet()) {
			final Object value = ((Map) o).get(key);
			switch (policy.isNeeded(value)) {
				case AS_LINK:
					if (value instanceof Identifiable) {
						jsonMap.put(this.toJSON(key, policy), makeLink((Identifiable) value, policy));
					}
					break;

				case YES:
					jsonMap.put(this.toJSON(key, policy), this.toJSON(value, policy));
					break;

				case NO:
					break;
			}
		}
		return json;
	}

	/**
	 * Converts an array to json representation
	 *
	 * @param o      - an array
	 * @param policy - policy to check what items are needed
	 *
	 * @return an JSONObject
	 *
	 * @throws SavingFailedException
	 */
	private JSONObject arrayToJSON(final Object o, final Policy policy) throws SavingFailedException {
		final JSONObject json = new JSONObject();
		//we do need to store additional information.
		json.put(ARRAY_SIZE, Array.getLength(o));
		json.put(ARRAY_COMPONENT, o.getClass().getComponentType().getName());
		json.put(TYPE_INFO, o.getClass().getName());

		final JSONArray array = new JSONArray();
		json.put(ARRAY, array);

		//and put all objects here
		for (int i = 0; i < Array.getLength(o); i++) {
			Object value = Array.get(o, i);
			switch (policy.isNeeded(value)) {
				case AS_LINK:
					if (value instanceof Identifiable) {
						array.add(makeLink((Identifiable) value, policy));
					}
					break;

				case YES:
					array.add(this.toJSON(value, policy));
					break;

				case NO:
					break;
			}
		}
		return json;
	}

	/// ---------------------------------------------------------------------------------------------- ///
	/// -------------------------------- Protected methods ------------------------------------------- ///
	/// ---------------------------------------------------------------------------------------------- ///

	/**
	 * Convert objects to JSON and write with <code>writer</code>
	 *
	 * @param writer     - the writer
	 * @param collection - the objects
	 * @param policy     - the save policy
	 *
	 * @throws SavingFailedException
	 */
	protected final void serialize(final Writer writer, final Collection<?> collection, Policy policy) throws SavingFailedException {
		try {
			//we can not allow null policy
			if (policy == null) {
				policy = new DefaultPolicy();
			}

			resetCache();

			final Iterator<?> it = collection.iterator();
			while (it.hasNext()) {
				final Object o = it.next();

				final Object data;
				// check that this object is needed
				switch (policy.isNeeded(o)) {
					case AS_LINK:
					case NO:
						if (o instanceof Identifiable) {
							data = makeLink((Identifiable) o, policy);
						} else {
							continue;
						}
						break;

					case YES:
						//get the object
						data = this.toJSON(o, policy);
						if (data instanceof JSONObject) {
							//don't save empty objects
							if (((JSONObject) data).isEmpty()) {
								continue;
							}
						}
						break;

					default:
						continue;
				}

				//make holder in case of simple types
				final JSONObject holder = new JSONObject();
				holder.put(DATA_TAG, data);

				//and write it to stream
				holder.writeJSONString(writer);
				writer.write(DELIMITER);
			}
			// flush data to ensure saving
			writer.flush();
		} catch (Exception e) {
			throw new SavingFailedException(e);
		}

	}

	/**
	 * Convert objects to JSON and write with <code>writer</code>
	 *
	 * @param writer     - the writer
	 * @param collection - the objects
	 *
	 * @throws SavingFailedException
	 */
	protected final void serialize(final Writer writer, final Collection<?> collection) throws SavingFailedException {
		serialize(writer, collection, null);
	}

	/**
	 * Serialize single object to writer
	 *
	 * @param writer - the writer
	 * @param o      - the object
	 * @param policy - the saving policy
	 *
	 * @throws SavingFailedException
	 */
	protected final void serialize(final Writer writer, final Object o, final Policy policy) throws SavingFailedException {
		serialize(writer, Collections.singletonList(o), policy);
	}

	/**
	 * Serialize single object to writer
	 *
	 * @param writer - the writer
	 * @param o      - the object
	 *
	 * @throws SavingFailedException
	 */
	protected final void serialize(final Writer writer, final Object o) throws SavingFailedException {
		serialize(writer, o, null);
	}

	/**
	 * Check if specified object is simple type
	 *
	 * @param o - an {@link Object}
	 *
	 * @return an {@link Boolean} value
	 */
	protected final static boolean isSimple(final Object o) {
		//FIXME: make it better
		// those damn simple types could make a great recursion hell, so they handled in special way
		return (!o.getClass().getName().replace("java.lang.", "").contains(".")) && !o.getClass().isArray();
	}

	/**
	 * Resets cache
	 */
	protected final void resetCache() {
		if (this.cache == null) {
			this.cache = new HashMap<Object, Identifiable>();
		} else {
			this.cache.clear();
		}
	}

	/**
	 * @param clazz - an {@link Class} object
	 *
	 * @return an {@link Field} object
	 */
	protected final static Field getNestedClassParentField(final Class<?> clazz) {
		Field parent = null;
		for (final Field f : clazz.getDeclaredFields()) {
			if (f.getType().equals(clazz.getEnclosingClass())) {
				parent = f;
				break;
			}
		}
		return parent;
	}

	/**
	 * Returns true if {@link Identifiable} interface implemented by this class
	 *
	 * @param clazz - an {@link Class} object
	 *
	 * @return an {@link Boolean} value
	 */
	protected final static boolean isImplementsIdentifiable(final Class<?> clazz) {
		//FIXME: find better way
		for (final Class<?> c : clazz.getInterfaces()) {
			if (c.equals(Identifiable.class)) { return true; }
		}
		return false;
	}

	/**
	 * Update cache, that object is saved, so future saving would be prevented
	 *
	 * @param o - the object
	 */
	protected final void saved(Identifiable o) {
		cache.put(o.getId(), o);
	}

	/**
	 * Ask cache if object already saved
	 *
	 * @param o - the object
	 *
	 * @return true if object saved
	 */
	protected final boolean isSaved(Identifiable o) {
		return cache.containsKey(o.getId());
	}

	/// ---------------------------------------------------------------------------------------------- ///
	/// --------------------------------- Abstract methods ------------------------------------------- ///
	/// ---------------------------------------------------------------------------------------------- ///


	/**
	 * Save objects
	 *
	 * @param objects - the objects container
	 * @param policy  - the save policy
	 *
	 * @throws SavingFailedException
	 */
	public abstract void save(final Collection<?> objects, final Policy policy) throws SavingFailedException;
}
