/******************************************************************************
 * Copyright (C) 2010-2011 Lastrix                                            *
 *                                                                            *
 * This file is part of JSONSerialization.                                    *
 *                                                                            *
 * JSONSerialization is free software: you can redistribute it and/or modify  *
 * it under the terms of the GNU General Public License as published by       *
 * the Free Software Foundation, either version 2 of the License, or          *
 * (at your option) any later version.                                        *
 *                                                                            *
 * JSONSerialization is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the               *
 * GNU General Public License for more details.                               *
 *                                                                            *
 * You should have received a copy of the GNU General Public License          *
 * along with JSONSerialization. If not, see <http://www.gnu.org/licenses/>.  *
 ******************************************************************************/

package org.lastrix.JSONSerialization.tool;

import org.lastrix.JSONSerialization.Policy;
import org.lastrix.JSONSerialization.exception.SavingFailedException;

import java.io.*;
import java.util.Collection;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

/**
 * Class to handle zip files.
 * Each call to #save(List, Policy) overwrites the content of file.
 *
 * @author: lastrix
 * Date: 8/10/11
 * Time: 6:30 PM
 */
public class ZIPJSONSaver extends AbstractJSONSaver {


	protected final File destination;

	/**
	 * Create default saver. You can not use #save() functions.
	 */
	public ZIPJSONSaver() {
		super();
		destination = null;
	}

	/**
	 * Create saver for specified file name
	 *
	 * @param fileName - the file name
	 *
	 * @throws IllegalArgumentException - if file can not be written
	 */
	public ZIPJSONSaver(final String fileName) throws IllegalArgumentException {
		this(new File(fileName));
	}

	/**
	 * Create saver for specified file
	 *
	 * @param destination - the file
	 *
	 * @throws IllegalArgumentException - if file can not be written
	 */
	public ZIPJSONSaver(final File destination) throws IllegalArgumentException {
		super();
		//if it doesn't exists
		if (!destination.exists()) {
			try {
				//try to create
				destination.createNewFile();
			} catch (IOException e) {
				//if not possible - send a message
				throw new IllegalArgumentException("Specified file '" + destination + "' can not be written.", e);
			}
		}
		//test access
		if (!destination.isFile() || !destination.canWrite()) {
			this.destination = null;
			throw new IllegalArgumentException("Specified file '" + destination + "' can not be written.");
		}
		this.destination = destination;
	}

	/**
	 * Save objects
	 *
	 * @param objects - the objects container
	 * @param policy  - the save policy
	 *
	 * @throws SavingFailedException
	 */
	@Override
	public void save(final Collection<?> objects, final Policy policy) throws SavingFailedException {
		if (destination == null) {
			throw new SavingFailedException("'Destination' is null.");
		}
		ZipOutputStream zos = null;
		try {
			zos = new ZipOutputStream(new FileOutputStream(destination));
		} catch (FileNotFoundException e) {
			throw new SavingFailedException(e);
		}

		try {
			zos.putNextEntry(new ZipEntry(ENTRY_NAME));
			BufferedOutputStream bos = new BufferedOutputStream(zos);
			OutputStreamWriter osw = new OutputStreamWriter(bos);
			serialize(osw, objects, policy);
			osw.flush();
			osw.close();
		} catch (IOException e) {
			throw new SavingFailedException(e);
		}
	}
}
