/******************************************************************************
 * Copyright (C) 2010-2011 Lastrix                                            *
 *                                                                            *
 * This file is part of JSONSerialization.                                    *
 *                                                                            *
 * JSONSerialization is free software: you can redistribute it and/or modify  *
 * it under the terms of the GNU General Public License as published by       *
 * the Free Software Foundation, either version 2 of the License, or          *
 * (at your option) any later version.                                        *
 *                                                                            *
 * JSONSerialization is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the               *
 * GNU General Public License for more details.                               *
 *                                                                            *
 * You should have received a copy of the GNU General Public License          *
 * along with JSONSerialization. If not, see <http://www.gnu.org/licenses/>.  *
 ******************************************************************************/

package org.lastrix.JSONSerialization.simple;

import junit.framework.TestCase;
import org.junit.Test;
import org.lastrix.JSONSerialization.Identifiable;
import org.lastrix.JSONSerialization.tool.PlainJSONLoader;
import org.lastrix.JSONSerialization.tool.PlainJSONSaver;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * @author: lastrix
 * Date: 8/3/11
 * Time: 10:56 AM
 */
public class SimpleTest extends TestCase {

	public static final String TEST1 = "Test1";

	public static class Pair<L, R> implements Identifiable<Integer> {
		private L   left;
		private R   right;
		private int id;

		private Pair() {
		}

		public Pair(final int id, final L left, final R right) {
			this.left = left;
			this.right = right;
			this.id = id;
		}


		public L getLeft() {
			return left;
		}

		public R getRight() {
			return right;
		}

		@Override
		public String toString() {
			return "<" + left +
			       ", " + right +
			       ">(" + id +
			       ')';
		}

		/**
		 * Return id of this object
		 *
		 * @return an <T> object
		 */
		@Override
		public Integer getId() {
			return id;
		}

		/**
		 * Set id
		 *
		 * @param id - an T value
		 */
		@Override
		public void setId(final Integer id) {
			this.id = id;
		}
	}

	public static class SimpleEntry implements Identifiable<Integer> {
		private Integer id;
		private String  value;

		protected SimpleEntry() {
		}

		public SimpleEntry(final Integer id, final String value) {
			this.id = id;
			this.value = value;
		}

		@Override
		public boolean equals(final Object obj) {
			if (this == obj) { return true; } else if (obj instanceof SimpleEntry) {
				SimpleEntry e = (SimpleEntry) obj;
				return e.getId() == getId() && e.getValue().equals(getValue());
			}
			return false;
		}

		@Override
		public int hashCode() {
			return id;
		}

		@Override
		protected Object clone() throws CloneNotSupportedException {
			return new SimpleEntry(id, value);
		}

		public String getValue() {
			return value;
		}

		public void setValue(final String value) {
			this.value = value;
		}

		/**
		 * Return id of this object
		 *
		 * @return an <T> object
		 */
		@Override
		public Integer getId() {
			return id;
		}

		/**
		 * Set id
		 *
		 * @param id - an T value
		 */
		@Override
		public void setId(final Integer id) {
			this.id = id;
		}

		@Override
		public String toString() {
			return "SimpleEntry{" +
			       "id=" + id +
			       ", value='" + value + '\'' +
			       '}';
		}
	}

	/**
	 * Test types saving and that its values are not changed
	 */
	@Test
	public void testNumbers() {
		final Object[] numbers = new Number[]{1, 2d, 3f, 4L};

		PlainJSONSaver saver = new PlainJSONSaver();
		PlainJSONLoader loader = new PlainJSONLoader();
		try {
			Number[] o = (Number[]) loader.deserializeObject(saver.serializeObject(numbers));
			assertTrue("Data corruption.", Arrays.equals(numbers, o));
		} catch (Exception e) {
			e.printStackTrace();
			fail("Exception occurred.");
		}
	}

	/**
	 * Test more complex types
	 */
	@Test
	public void testSimples() {

		final SimpleEntry[] entries = new SimpleEntry[]{
		                                               new SimpleEntry(1, TEST1),
		                                               new SimpleEntry(2, TEST1),
		                                               new SimpleEntry(3, TEST1),
		                                               new SimpleEntry(4, TEST1),
		                                               new SimpleEntry(5, TEST1)
		};


		PlainJSONSaver saver = new PlainJSONSaver();
		PlainJSONLoader loader = new PlainJSONLoader();

		try {
			SimpleEntry[] o = (SimpleEntry[]) loader.deserializeObject(saver.serializeObject(entries));

			List<Integer> ids = new ArrayList<Integer>();
			for (Object e : o) {
				if (e instanceof SimpleEntry) {
					if (ids.contains(((SimpleEntry) e).getId())) {
						fail("Duplicate entry found.");
					}
					if (!TEST1.equals(((SimpleEntry) e).getValue())) {
						fail(String.format("Entry value is not '%s'", TEST1));
					}
				} else {
					fail("An incompatible object found.");
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
			fail("Exception occured.");
		}

	}

	@Test
	public void testSimplePairs() {
		final int offset = 100;
		List list = new ArrayList();

		PlainJSONSaver saver = new PlainJSONSaver();
		PlainJSONLoader loader = new PlainJSONLoader();

		for (int id = 1; id < 5; id++) {
			list.add(new Pair<String, SimpleEntry>(id, TEST1, new SimpleEntry(offset + id, TEST1)));
		}
		try {
			list = (List) loader.deserializeObject(saver.serializeObject(list));
			for (Object e : list) {
				if (e instanceof Pair) {
					if (((Pair) e).getRight() instanceof SimpleEntry) {
						if (((SimpleEntry) ((Pair) e).getRight()).getId() - offset != ((Pair) e).getId()) {
							fail("Corrupted data " + ((SimpleEntry) ((Pair) e).getRight()).getId() + " " + ((Pair) e).getId());
						}
					} else {
						fail("An incompatible object found2 (" + ((Pair) e).getRight().getClass() + ").");
					}
				} else if (e instanceof SimpleEntry) {
					//not so bad
				} else {
					fail("An incompatible object found(" + e.getClass() + ").");
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
			fail("Exception occured.");
		}
	}
}
